/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common.su;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.objectweb.petals.common.descriptor.Consumes;
import org.objectweb.petals.common.descriptor.JBIDescriptor;
import org.objectweb.petals.common.descriptor.JBIDescriptorBuilder;
import org.objectweb.petals.common.descriptor.JBIDescriptorException;
import org.objectweb.petals.common.descriptor.Provides;
import org.objectweb.petals.component.common.AbstractComponent;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.w3c.dom.Document;

/**
 * A simple ServiceUnitManager. It processes service unit jbi descriptor : <br/>
 * Provides nodes : new endpoints registered in JBI container<br/> Consumes
 * nodes : link addresses to existing JBI endpoints
 *
 * @author ofabre,alouis
 *
 */
public class SimpleServiceUnitManager implements ServiceUnitManager {

    private static final String JBI_XML = "jbi.xml";

    private static final String META_INF = "META-INF";

    private Map<ServiceEndpoint, ServiceUnitDataHandler> suHandlerCache;

    /**
     * The component's context
     */
    private final ComponentContext context;

    /**
     * The component's Logger
     */
    private Logger logger; // NOPMD by gblondelle

    private final Map<String, ServiceUnitDataHandler> serviceUnitDatas;

    private final AbstractComponent component;

    /**
     * Creates a new instance of {@link SimpleServiceUnitManager}
     *
     * @param engine
     * @param context
     * @param logger
     */
    public SimpleServiceUnitManager(AbstractComponent engine,
        ComponentContext context, Logger logger) {
        super();
        this.context = context;
        this.logger = logger;
        this.component = engine;
        serviceUnitDatas = new HashMap<String, ServiceUnitDataHandler>();
        suHandlerCache = new HashMap<ServiceEndpoint, ServiceUnitDataHandler>();
    }

    /**
     *
     */
    public String deploy(final String serviceUnitName, final String suRootPath)
        throws DeploymentException {
        logger.log(Level.FINE, "deploy  serviceUnitName  " + serviceUnitName
            + "serviceUnitRootPath" + suRootPath);
        if (serviceUnitDatas.get(serviceUnitName) == null) {
            ServiceUnitDataHandler handler = new ServiceUnitDataHandler();
            handler.setInstallRoot(suRootPath);
            handler.setName(serviceUnitName);
            File jbiXmlFile = new File(suRootPath + File.separator + META_INF
                + File.separator + JBI_XML);
            JBIDescriptor descriptor = null;
            try {
                descriptor = new JBIDescriptorBuilder(jbiXmlFile.toURI(),
                    logger).build();
            } catch (JBIDescriptorException e) {
                String msg = "Bad JBI Descriptor";
                logger.log(Level.SEVERE, msg);
                throw new DeploymentException(msg, e);
            }
            if (descriptor != null) {
                handler.setDescriptor(descriptor);

                try {
                    processConsumesNodes(handler);
                } catch (JBIException ex) {
                    String msg = "Failed to process consumes nodes";
                    logger.log(Level.SEVERE, msg, ex);
                    throw new DeploymentException(msg, ex);
                }

                serviceUnitDatas.put(serviceUnitName, handler);

                if (component.getServiceUnitListener() != null) {
                    try {
                        component.getServiceUnitListener().onSUDeployed(
                            serviceUnitName, suRootPath, descriptor);
                    } catch (PEtALSComponentSDKException e) {
                        // clean cache
                        serviceUnitDatas.remove(serviceUnitName);

                        throw new DeploymentException(
                            "The ServiceUnit listener failed to process.", e);
                    }
                }
            }
        } else {
            throw new DeploymentException(
                "A service unit with the given name is already registered");
        }
        String result = ManagementMessageUtil.getComponentTaskResult(context
            .getComponentName(), "deploy",
            ManagementMessageUtil.TASK_RESULT_SUCCESS);
        return result;
    }

    /**
     * Retrieve the Consumes node linked to the given address
     *
     * @param address
     *            address of the search Consumes node
     * @return a {@link Consumes} element
     */
    public Consumes getConsumesFromAddress(final String address) {
        Consumes result = null;
        for (Iterator iter = serviceUnitDatas.values().iterator(); iter
            .hasNext()
            && result == null;) {
            ServiceUnitDataHandler handler = (ServiceUnitDataHandler) iter
                .next();
            Map<String, Consumes> addressToConsumes = handler
                .getAddressToConsumes();
            result = addressToConsumes.get(address);
        }
        return result;
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     *
     * @param serviceEndpoint
     *            the previously activated jbi Endpoint
     * @return the {@link Provides} element
     */
    public Provides getProvidesFromEndpoint(
        final ServiceEndpoint serviceEndpoint) {
        Provides result = null;
        for (Iterator iter = serviceUnitDatas.values().iterator(); iter
            .hasNext()
            && result == null;) {
            ServiceUnitDataHandler handler = (ServiceUnitDataHandler) iter
                .next();
            Map<ServiceEndpoint, Provides> epJBIDescs = handler.getEpJBIDesc();
            result = epJBIDescs.get(serviceEndpoint);
        }
        return result;
    }

    /**
     * Retrieve the Service description (as Dom Document) of the given jbi
     * Endpoint
     *
     * @param serviceEndpoint
     *            the jbi Endpoint
     * @return the service description as Dom Document
     */
    public Document getServiceDescription(final ServiceEndpoint serviceEndpoint) {
        Document result = null;
        for (Iterator iter = serviceUnitDatas.values().iterator(); iter
            .hasNext()
            && result == null;) {
            ServiceUnitDataHandler handler = (ServiceUnitDataHandler) iter
                .next();
            Map<ServiceEndpoint, Document> epServDescs = handler
                .getEpServiceDesc();
            result = epServDescs.get(serviceEndpoint);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String, java.lang.String)
     */
    public void init(final String serviceUnitName, final String suRootPath)
        throws DeploymentException {
        logger.log(Level.FINE, "init  serviceUnitName  " + serviceUnitName
            + "  serviceUnitRootPath " + suRootPath);
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    public void shutDown(final String serviceUnitName)
        throws DeploymentException {
        logger.log(Level.FINE, "shutdown  serviceUnitName  " + serviceUnitName);
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     */
    public void start(final String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);
        ServiceUnitDataHandler handler = serviceUnitDatas.get(serviceUnitName);
        try {
            processProvidesNodes(handler);
        } catch (JBIException ex) {
            String msg = "Failed to start service unit : " + serviceUnitName;
            logger.log(Level.SEVERE, msg, ex);
            throw new DeploymentException(msg, ex);
        }
        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUStarted(serviceUnitName);
            } catch (PEtALSComponentSDKException e) {
                throw new DeploymentException(
                    "The ServiceUnit listener failed to process.", e);
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     */
    public void stop(final String serviceUnitName) throws DeploymentException {
        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUStopped(serviceUnitName);
            } catch (PEtALSComponentSDKException e) {
                throw new DeploymentException(
                    "The ServiceUnit listener failed to process.", e);
            }
        }
        try {
            ServiceUnitDataHandler handler = serviceUnitDatas
                .get(serviceUnitName);
            List<ServiceEndpoint> endpoints = getServiceUnitEndpoints(serviceUnitName);
            for (ServiceEndpoint endpoint : endpoints) {
                context.deactivateEndpoint(endpoint);
                handler.removeEndpoint(endpoint);
            }

        } catch (JBIException e) {
            throw new DeploymentException("Deactivate endpoint failed", e);
        }
        logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);
    }

    /*
     * (non-Javadoc)
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String, java.lang.String)
     */
    public String undeploy(final String serviceUnitName,
        final String serviceUnitRootPath) throws DeploymentException {
        ServiceUnitDataHandler handler = serviceUnitDatas.get(serviceUnitName);

        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUUndeployed(
                    serviceUnitName, serviceUnitRootPath,
                    handler.getDescriptor());
            } catch (PEtALSComponentSDKException e) {
                throw new DeploymentException(
                    "The ServiceUnit listener failed to process.", e);
            }
        }

        // clean the chache
        for (ServiceEndpoint se : suHandlerCache.keySet()) {
            if (suHandlerCache.get(se).equals(handler)) {
                suHandlerCache.remove(se);
            }
        }

        serviceUnitDatas.remove(serviceUnitName);
        return ManagementMessageUtil.getComponentTaskResult(context
            .getComponentName(), "undeploy",
            ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    /**
     * Search the wsdl description of the activated endpoint in the Provides
     * extensions. If no wsdl description can be found, it creates a light
     * weight Wsdl description from Provides Interface, Service and Endpoint
     * names.
     *
     * @param installRoot
     *            the SU install root
     * @param provides
     *            the Provices node
     * @return the WSDL description as Dom Document
     */
    @SuppressWarnings("unchecked")
    protected Document getEnpointWSDLDesc(String installRoot, Provides provides) {

        Document wsdlDesc = null;

        String wsdlFileLocation = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(provides.getExtensions(),
                PetalsExtensionsUtil.WSDL_LOCATION);

        if (wsdlFileLocation != null) {

            if (wsdlFileLocation.toLowerCase().startsWith("http:")
                || wsdlFileLocation.toLowerCase().startsWith("https:")
                || wsdlFileLocation.toLowerCase().startsWith("file:")) {

                try {
                    URL url = new URL(wsdlFileLocation);
                    wsdlDesc = WSDLHelper.createDocumentFromWSDL(url);
                } catch (MalformedURLException e) {
                    // Nothing to do
                }

            } else {
                // get local wsdl (relative or absolute)
                if (!wsdlFileLocation.startsWith(File.separator)) {
                    wsdlFileLocation = File.separator + wsdlFileLocation;
                }

                File wsdlFile = new File(installRoot + wsdlFileLocation);
                wsdlDesc = WSDLHelper.createDocumentFromWSDL(wsdlFile);
            }
        }

        // if the wsdl description is null, generate it
        if (wsdlDesc == null) {
            wsdlDesc = WSDLHelper.createLightWSDL20(
                provides.getInterfaceName(), provides.getServiceName(),
                provides.getEndpointName());
        }

        return wsdlDesc;
    }

    /**
     *
     * @param handler
     * @param consumes
     * @throws JBIException
     */
    protected void processConsumesNode(final ServiceUnitDataHandler handler,
        final Consumes consumes) throws JBIException {
        String address = PetalsExtensionsUtil
            .extractValueFromKeyValueExtension(consumes.getExtensions(),
                PetalsExtensionsUtil.ADDRESS);

        if (address != null) {
            handler.addAddressToConsumes(address, consumes);
        }
    }

    /**
     *
     * @param handler
     * @throws JBIException
     */
    protected void processConsumesNodes(final ServiceUnitDataHandler handler)
        throws JBIException {
        JBIDescriptor descriptor = handler.getDescriptor();
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            processConsumesNode(handler, consumes);
        }
    }

    /**
     *
     * @param handler
     * @param provides
     * @return
     * @throws JBIException
     */
    protected ServiceEndpoint processProvidesNode(
        final ServiceUnitDataHandler handler, final Provides provides)
        throws JBIException {
        ServiceEndpoint ep = context.activateEndpoint(
            provides.getServiceName(), provides.getEndpointName());
        handler.addEndpointDescription(ep, getEnpointWSDLDesc(handler
            .getInstallRoot(), provides));
        handler.addJBIDescription(ep, provides);
        return ep;
    }

    /**
     *
     * @param handler
     * @throws JBIException
     */
    protected void processProvidesNodes(final ServiceUnitDataHandler handler)
        throws JBIException {
        JBIDescriptor descriptor = handler.getDescriptor();
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            processProvidesNode(handler, provides);
        }
    }

    /**
     *
     * @param serviceUnitName
     * @return
     */
    private List<ServiceEndpoint> getServiceUnitEndpoints(
        final String serviceUnitName) {
        return new ArrayList<ServiceEndpoint>(serviceUnitDatas.get(
            serviceUnitName).getEpServiceDesc().keySet());
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s as a Map
     *
     * @return all {@link ServiceUnitDataHandler}s as a Map
     */
    public Map<String, ServiceUnitDataHandler> getServiceUnitDatas() {
        return serviceUnitDatas;
    }

    /**
     * Return the component Logger
     *
     * @return the component Logger
     */
    public Logger getLogger() {
        return logger;
    }

    /**
     * Retrieve the install root of the Service Unit that activated the given
     * endpoint
     *
     * @param ep
     *            the previously activated endpoint
     * @return the Service unit install root
     */
    public String getSUInstallRootForActivatedEp(ServiceEndpoint ep) {
        ServiceUnitDataHandler handler = getSuDataHandlerForEp(ep);
        return handler.getInstallRoot();
    }

    /**
     * Retrieve the data handler of the Service Unit that activated the given
     * endpoint
     *
     * @param ep
     *            the previously activated endpoint
     * @return the {@link ServiceUnitDataHandler}
     */
    public ServiceUnitDataHandler getSuDataHandlerForEp(ServiceEndpoint ep) {

        ServiceUnitDataHandler result = suHandlerCache.get(ep);

        if (result == null) {
            for (Iterator iter = getServiceUnitDatas().values().iterator(); iter
                .hasNext()
                && result == null;) {
                ServiceUnitDataHandler element = (ServiceUnitDataHandler) iter
                    .next();
                for (Iterator iterator2 = element.getEpJBIDesc().keySet()
                    .iterator(); iterator2.hasNext() && result == null;) {
                    ServiceEndpoint element2 = (ServiceEndpoint) iterator2
                        .next();
                    if (element2.equals(ep)) {
                        result = element;
                    }
                }
            }
        }
        return result;
    }
}
